using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using UnityEngine;

namespace CheesyFX
{
    public static class KeyInput
    {
        public static JSONStorableFloat doubleTapTimeout = new JSONStorableFloat("Double Tap Timeout", .3f, .1f, 1f);
        public static Action<int, KeyCode> onSingle;
        public static Action<int> onDouble;
        public static Action<int> onDuringSingleStart;
        public static Action<int> duringSingleDown;
        public static Action<int> onDuringSingleEnd;
        public static Action<int> onDuringDoubleStart;
        public static Action<int> duringDoubleDown;
        public static Action<int> onDuringDoubleEnd;

        // private static KeyCode[] modifiers = 
        // {
        //     KeyCode.LeftControl,
        //     KeyCode.RightControl,
        //     KeyCode.LeftShift,
        //     KeyCode.RightShift,
        //     KeyCode.LeftAlt,
        //     KeyCode.RightAlt
        // };
        public static KeyCode[] modifiers;
        public static JSONStorableBool[] keysHeldDown;
        public static int[] tapCounts;

        public static void SetUpKeys(string[] keyNames)
        {
            int keyNumber = keyNames.Length;
            keysHeldDown = new JSONStorableBool[keyNumber];
            tapCounts = new int[keyNumber];
            for (int i = 0; i < keyNumber; i++)
            {
                int id = i;
                keysHeldDown[i] = new JSONStorableBool(keyNames[i], false, x =>
                {
                    if (x) OnKeyDown(id);
                });
                tapCounts[i] = 0;
            }
        }
        
        public static void SetUpKeys(int keyCount)
        {
            keysHeldDown = new JSONStorableBool[keyCount];
            tapCounts = new int[keyCount];
            for (int i = 0; i < keyCount; i++)
            {
                int id = i;
                keysHeldDown[i] = new JSONStorableBool($"Key{i}", false, x =>
                {
                    if (x) OnKeyDown(id);
                });
                tapCounts[i] = 0;
            }
        }

        public static void OnKeyDown(int keyId, bool fromKeybindings = false)
        {
            if (fromKeybindings && !LightMeUp.useKeybindings.val) return;
            if (tapCounts[keyId] == 0)
            {
                KeyCode modifier = modifiers.FirstOrDefault(Input.GetKey);
                if (modifier != KeyCode.None)
                {
                    onSingle(keyId, modifier);
                }
                else
                {
                    tapCounts[keyId] = 1;
                    keysHeldDown[keyId].val = true;
                    TapInit(keyId).Start();
                }
            }
            else
            {
                keysHeldDown[keyId].val = true;
                tapCounts[keyId] += 1;
            }
            
        }

        static IEnumerator TapInit(int keyId)
        {
            yield return new WaitForSeconds(doubleTapTimeout.val);
            int tapCount = tapCounts[keyId];
            if (keysHeldDown[keyId].val)
            {
                if (tapCount == 1)
                {
                    onDuringSingleStart?.Invoke(keyId);
                    if(duringSingleDown != null) DuringSingleDown(keyId).Start();
                }

                if (tapCount > 1)
                {
                    onDuringDoubleStart?.Invoke(keyId);
                    if(duringDoubleDown != null) DuringDoubleDown(keyId).Start();
                }
            }
            else
            {
                if (tapCount == 1 && onSingle != null) onSingle(keyId, KeyCode.None);
                if (tapCount > 1 && onDouble != null) onDouble(keyId);
                tapCounts[keyId] = 0;
            }
        }

        private static IEnumerator DuringSingleDown(int keyId)
        {
            while (keysHeldDown[keyId].val)
            {
                // "During Single Down".Print();
                duringSingleDown(keyId);
                yield return new WaitForFixedUpdate();
            }
            onDuringSingleEnd?.Invoke(keyId);
            tapCounts[keyId] = 0;
        }
        
        private static IEnumerator DuringDoubleDown(int keyId)
        {
            while (keysHeldDown[keyId].val)
            {
                // "During Double Down".Print();
                duringDoubleDown(keyId);
                yield return new WaitForFixedUpdate();
            }
            onDuringDoubleEnd?.Invoke(keyId);
            tapCounts[keyId] = 0;
        }
    }
}