using System;
using System.Linq;
using System.Collections.Generic;
using MacGruber;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using UnityEngine;
using UnityEngine.UI;


namespace CheesyFX
{
    public partial class LightMeUp : MVRScript
    {
        private static bool _toggleState = true;
        public static List<HeadLight> headLights = new List<HeadLight>();
        public static Transform cameraTransform;
        
        private static JSONStorableStringChooser lightChooser;
        public static JSONStorableBool editSymmetrical;
        private static JSONStorableBool autoFocus = new JSONStorableBool("Auto Focus Side Lights", false);
        
        private static JSONStorableStringChooser leftKeyChooser;
        private static JSONStorableStringChooser centerKeyChooser;
        private static JSONStorableStringChooser rightKeyChooser;
        private static JSONStorableStringChooser allKeyChooser;
        private static JSONStorableStringChooser mod0Chooser;
        private static JSONStorableStringChooser mod1Chooser;
        private static JSONStorableStringChooser mod2Chooser;
        
        public static JSONStorableBool useKeybindings;
        

        private bool sideLightEnabled => headLights[0].enabled.val || headLights[2].enabled.val;
        
        private static RectTransform versionRectTransform;
        private static RectTransform buttonsTransform;
        private static RectTransform toggleButtonTransform;
        
        private ColorBlock activeColorBlock;
        public static ColorBlock disabledColorBlock;

        // private static KeyCode toggleAll = KeyCode.KeypadPlus;
        // private static KeyCode toggleLeft = KeyCode.KeypadDivide;
        // private static KeyCode toggleCenter = KeyCode.KeypadMultiply;
        // private static KeyCode toggleRight = KeyCode.KeypadMinus;

        private static KeyCode[] toggleKeys = {
            KeyCode.KeypadDivide,
            KeyCode.KeypadMultiply,
            KeyCode.KeypadMinus,
            KeyCode.KeypadPlus,
        };
        private static KeyCode[] modifiers = {
            KeyCode.RightControl,
            KeyCode.RightShift,
            KeyCode.LeftAlt
        };
        public static bool disableSymmetry;

        private static JSONStorableFloat IntensityRate = new JSONStorableFloat("Intensity Rate", .002f, .001f, .05f);

        private List<object> UIElements = new List<object>();
        private List<object> LightUIElements = new List<object>();

        public static Color[] defaultColors = 
        {
            new Color(1f, 0.89f, 0.78f),
            new Color(1f,1f,1f),
            new Color(0.78f, 0.98f, 1f),
        };
        
        private static Dictionary<Atom, bool> sceneLights = new Dictionary<Atom, bool>();
        private static bool isSceneLit = true;
        public static bool isOnVR;
        private static bool keybindingsPresent;

        private UIDynamicTextInfo lightInfo;

        public override void Init()
        {
            // SuperController.singleton.ClearMessages();
            // PrintHelp();
            MacGruber.Utils.OnInitUI(CreateUIElement);;
            SetCameraRig();
            disabledColorBlock = ColorBlock.defaultColorBlock;
            disabledColorBlock.normalColor *= .5f;
            disabledColorBlock.highlightedColor *= .5f;
            disabledColorBlock.pressedColor *= .5f;
            headLights = new List<HeadLight>()
            {
                new HeadLight(this, 0),
                new HeadLight(this, 1),
                new HeadLight(this, 2)
            };
            
            CreateMainHUD();
            
            lightChooser = new JSONStorableStringChooser(
                "LightChooser",
                headLights.Select(x => x.name).ToList(),
                "Center",
                "Light",
                CreateLightUI
                );
            List<string> keyOptions = Enum.GetNames(typeof(KeyCode)).ToList();
            leftKeyChooser = new JSONStorableStringChooser("Key0", keyOptions,
                toggleKeys[0].ToString(), "Key Left", SetKey);
            rightKeyChooser = new JSONStorableStringChooser("Key2", keyOptions,
                toggleKeys[2].ToString(), "Key Right", SetKey);
            centerKeyChooser = new JSONStorableStringChooser("Key1", keyOptions,
                toggleKeys[1].ToString(), "Key Center", SetKey); 
            allKeyChooser = new JSONStorableStringChooser("Key3", keyOptions,
                toggleKeys[3].ToString(), "Key All", SetKey);
            
            mod0Chooser = new JSONStorableStringChooser("Mod0", keyOptions,
                modifiers[0].ToString(), "Modifier0", SetModifier);
            mod1Chooser = new JSONStorableStringChooser("Mod1", keyOptions,
                modifiers[1].ToString(), "Modifier1", SetModifier);
            mod2Chooser = new JSONStorableStringChooser("Mod2", keyOptions,
                modifiers[2].ToString(), "Modifier2", SetModifier);

            useKeybindings = new JSONStorableBool("Use AcidBubbles Keybindings", false, UseKeybindingsCallback);
            editSymmetrical = new JSONStorableBool("Edit Symmetrical (except color)", true);
            
            Toggle();
            SetupKeyInput();
            SuperController.singleton.BroadcastMessage("OnActionsProviderAvailable", this, SendMessageOptions.DontRequireReceiver);
            CreateUI();
        }
        
        
        public void OnBindingsListRequested(List<object> bindings)
        {
            bindings.Add(new Dictionary<string, string>
            {
                { "CheesyFX", "LightMeUp" }
            });

            for (int i = 0; i < toggleKeys.Length; i++)
            {
                bindings.Add(KeyInput.keysHeldDown[i]);
            }

            bindings.Add(new JSONStorableAction("All Reset Intensity", () => ResetIntensity(3, true)));
            bindings.Add(new JSONStorableAction("Left Reset Intensity", () => ResetIntensity(0, true)));
            bindings.Add(new JSONStorableAction("Center Reset Intensity", () => ResetIntensity(1, true)));
            bindings.Add(new JSONStorableAction("Right Reset Intensity", () => ResetIntensity(2, true)));
            
            bindings.Add(new JSONStorableAction("All Toggle Shadows", () => ToggleShadows(3, true)));
            bindings.Add(new JSONStorableAction("Left Toggle Shadows", () => ToggleShadows(0, true)));
            bindings.Add(new JSONStorableAction("Center Toggle Shadows", () => ToggleShadows(1, true)));
            bindings.Add(new JSONStorableAction("Right Toggle Shadows", () => ToggleShadows(2, true)));
            
            bindings.Add(new JSONStorableAction("Toggle Scene Lights", () => ToggleSceneLighting(true)));
            bindings.Add(new JSONStorableAction("Swap L/R colors", () => SwapLeftRightColor(true)));
            
            bindings.Add(new JSONStorableAction("Left Intensity + 0.1", () => headLights[0].intensity.val+=.1f));
            bindings.Add(new JSONStorableAction("Center Intensity + 0.1", () => headLights[1].intensity.val+=.1f));
            bindings.Add(new JSONStorableAction("Right Intensity + 0.1", () => headLights[2].intensity.val+=.1f));
            bindings.Add(new JSONStorableAction("All Intensity + 0.1", () => headLights.ForEach(x=>x.intensity.val+=.1f)));
            
            bindings.Add(new JSONStorableAction("Left Intensity - 0.1", () => headLights[0].intensity.val-=.1f));
            bindings.Add(new JSONStorableAction("Center Intensity - 0.1", () => headLights[1].intensity.val-=.1f));
            bindings.Add(new JSONStorableAction("Right Intensity - 0.1", () => headLights[2].intensity.val-=.1f));
            bindings.Add(new JSONStorableAction("All Intensity - 0.1", () => headLights.ForEach(x=>x.intensity.val-=.1f)));

            keybindingsPresent = true;
        }

        private void PrintHelp()
        {
            string help = "LightMeUp shortcuts:\n\n" +
                          "Main Keys:\n" +
                          $"Left:\t\t\t\t\t\t\t\t\t\t\t{toggleKeys[0]}\n" +
                          $"Center:\t\t\t\t\t\t\t\t\t\t{toggleKeys[1]}\n" +
                          $"Right:\t\t\t\t\t\t\t\t\t\t{toggleKeys[2]}\n" +
                          $"All:\t\t\t\t\t\t\t\t\t\t\t{toggleKeys[3]}\n\n" +
                          "Tap & Hold:\t\t\t\t\t\t\t\tincrease intensity\n" +
                          "Double Tap & Hold:\t\t\t\t\tdecrease intensity\n\n" +
                          "Modifiers:\n" +
                          $"Reset Intensity:\t\t\t\t\t\t{modifiers[0]}+<any>\n" +
                          $"Toggle Shadows:\t\t\t\t\t\t{modifiers[1]}+<any>\n" +
                          $"SceneLights / Swap color:\t\t{modifiers[2]}+<any>\n";
            help.Print();
        }

        private static void SetupKeyInput()
        {
            KeyInput.onSingle = OnSingleTap;
            KeyInput.onDuringSingleStart = key => disableSymmetry = true;
            KeyInput.onDuringSingleEnd = key => disableSymmetry = false;
            KeyInput.onDuringDoubleStart = key => disableSymmetry = true;
            KeyInput.onDuringDoubleEnd = key => disableSymmetry = false;
            KeyInput.duringSingleDown = key => AdjustIntensity(key, 1f);
            KeyInput.duringDoubleDown = key => AdjustIntensity(key, -1f);

            KeyInput.SetUpKeys(new []
            {
                "Left (Toggle & Intensity)",
                "Center (Toggle & Intensity)",
                "Right (Toggle & Intensity)",
                "All (Toggle & Intensity)"
            });
            KeyInput.modifiers = modifiers;
        }

        private static void SetKey(JSONStorableStringChooser jString)
        {
            int id = int.Parse(jString.name[3].ToString());
            toggleKeys[id] = (KeyCode)Enum.Parse(typeof(KeyCode), jString.val);
        }
        
        private static void SetModifier(JSONStorableStringChooser jString)
        {
            int id = int.Parse(jString.name[3].ToString());
            KeyInput.modifiers[id] = (KeyCode)Enum.Parse(typeof(KeyCode), jString.val);
        }

        private static void OnSingleTap(int keyId, KeyCode modifier)
        {
            if(modifier == KeyCode.LeftAlt)
            {
                if(keyId==3) ToggleSceneLighting();
                else SwapLeftRightColor();
            }
            else
            {
                if(modifier == KeyCode.LeftControl || modifier == KeyCode.RightControl) ResetIntensity(keyId);
                else if (modifier == KeyCode.LeftShift || modifier == KeyCode.RightShift) ToggleShadows(keyId);
                else Toggle(keyId);
            }
        }

        private void UseKeybindingsCallback(bool b)
        {
            if (b && !keybindingsPresent)
            {
                SuperController.LogError("LightMeUp: Plugin AcidBubbles.Keybindings not found.\n" +
                                         "Add Keybindings as a session plugin first, then reload this plugin.");
                useKeybindings.valNoCallback = false;
            }
        }

        private void SetCameraRig()
        {
            SuperController singleton = SuperController.singleton;
            if (singleton.isOpenVR)
            {
                cameraTransform = singleton.ViveCenterCamera.transform;
                isOnVR = true;
            }
            else if (singleton.isOVR)
            {
                cameraTransform = singleton.OVRCenterCamera.transform;
                isOnVR = true;
            }
            else if (singleton.IsMonitorOnly)
            {
                cameraTransform = singleton.MonitorCenterCamera.transform;
            }
        }

        private void CreateLightUI(JSONStorableStringChooser jString)
        {
            lightInfo.text.text = jString.val + " Light Settings";
            MacGruber.Utils.RemoveUIElements(this, LightUIElements);
            LightUIElements = headLights[lightChooser.choices.IndexOf(jString.val)].CreateUI();
        }
        

        public void Update()
        {
            if(!useKeybindings.val)
            {
                for (int i=0; i< toggleKeys.Length; i++)
                {
                    if (Input.GetKeyDown(toggleKeys[i])) KeyInput.keysHeldDown[i].val = true;
                    else if (Input.GetKeyUp(toggleKeys[i])) KeyInput.keysHeldDown[i].val = false;
                }
            }
            if (!autoFocus.val || !_toggleState || !sideLightEnabled) return;
            RaycastHit hit;
            Ray ray = new Ray(cameraTransform.position, cameraTransform.forward);
            if(Physics.Raycast(ray, out hit))
            {
                Vector3 focusPoint = hit.point;
                headLights[0].UpdateRotation(focusPoint);
                headLights[2].UpdateRotation(focusPoint);
            }
        }

        private static void Toggle(int id = 3)
        {
            if(id == 3 || !_toggleState)
            {
                toggleButtonTransform.gameObject.SetActive(_toggleState);
                _toggleState = !_toggleState;
                headLights.ForEach(x => x.EnabledCallback(_toggleState && x.enabled.val));
            }
            else
            {
                headLights[id].Toggle();
                if (_toggleState) return;
                toggleButtonTransform.gameObject.SetActive(false);
                _toggleState = true;
            }
        }
        
        private static void AdjustIntensity(int keyId, float direction)
        {
            float increment = IntensityRate.val * direction;
            if (keyId < 3) headLights[keyId].intensity.val += increment;
            else headLights.ForEach(x => x.intensity.val += increment);
        }

        private static void ResetIntensity(int id, bool fromKeybindings = false)
        {
            if (fromKeybindings && !useKeybindings.val) return;
            if(id == 3) headLights.ForEach(x => x.intensity.SetValToDefault());
            else
            {
                bool sym = editSymmetrical.val;
                editSymmetrical.val = false;
                headLights[id].intensity.SetValToDefault();
                editSymmetrical.val = sym;
            }
        }

        private static void SwapLeftRightColor(bool fromKeybindings = false)
        {
            if (fromKeybindings && !useKeybindings.val) return;
            HSVColor leftColor = headLights[0].color.val;
            headLights[0].color.val = headLights[2].color.val;
            headLights[2].color.val = leftColor;
        }

        private static void ToggleShadows(int id, bool fromKeybindings = false)
        {
            if (fromKeybindings && !useKeybindings.val) return;
            if (id == 3)
            {
                headLights.ForEach(x => x.castShadows.val = !x.castShadows.val);
            }
            else headLights[id].castShadows.val = !headLights[id].castShadows.val;
        }

        private static void ToggleSceneLighting(bool fromKeybindings = false)
        {
            if (fromKeybindings && !useKeybindings.val) return;
            if(isSceneLit) sceneLights.Clear();
            foreach (var atom in SuperController.singleton.GetAtoms())
            {
                if (atom.type == "InvisibleLight")
                {
                    JSONStorableBool on = atom.GetStorableByID("Light").GetBoolJSONParam("on");
                    if (isSceneLit)
                    {
                        sceneLights.Add(atom, on.val);
                        on.val = false;
                    }
                    else on.val = sceneLights[atom];
                }
            }
            isSceneLit = !isSceneLit;
        }

        public void OnEnable()
        {
            headLights.ForEach(x => x.enabled.val = true); 
        }

        public void OnDisable()
        {
            headLights.ForEach(x => x.enabled.val = false);
            if(!isSceneLit) ToggleSceneLighting();
        }
        
        public void OnDestroy()
        {
            headLights.ForEach(x => x.Destroy());
            versionRectTransform.anchorMax = new Vector2(.37f, 0f);
            Destroy(toggleButtonTransform.gameObject);
            if(!isSceneLit) ToggleSceneLighting();
            SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);
        }
        
        private void CreateUI()
        {
            MacGruber.Utils.RemoveUIElements(this, UIElements);
            UIElements.Clear();
            
            UIDynamic spacer;
            // UIDynamic spacer = CreateSpacer(true);
            // spacer.height = 10f;
            
            UIElements.Add(CreateToggle(editSymmetrical, false));
            UIElements.Add(CreateToggle(useKeybindings, false));
            UIElements.Add(CreateToggle(autoFocus, true));
            
            
            UIElements.Add(MacGruber.Utils.SetupTwinButton(this, "Swap L/R colors", () => SwapLeftRightColor(), "ToggleSceneLights", () => ToggleSceneLighting(), true));
            // MacGruber.Utils.SetupButton(this, "Swap L/R colors", SwapLeftRightColor, false);
            // MacGruber.Utils.SetupButton(this, "Toggle Scene Lights", ToggleSceneLighting, true);
            // spacer = CreateSpacer(true);
            // spacer.height = 45f;
            
            lightInfo = MacGruber.Utils.SetupInfoOneLine(this, "Center Light", false);
            lightInfo.height = 50f;
            lightInfo.text.fontSize = 32;
            lightInfo.text.alignment = TextAnchor.MiddleCenter;
            UIElements.Add(lightInfo);
            UIElements.Add(MacGruber.Utils.SetupTwinButton(this, "Advanced Options", CreateAdvancedUI, "Print Help", PrintHelp, true));
            // UIElements.Add(MacGruber.Utils.SetupButton(this, "Advanced Options", CreateAdvancedUI, true));
            // spacer = CreateSpacer(true);
            // spacer.height = 50f;
            // UIElements.Add(spacer);
            UIElements.Add(CreatePopup(lightChooser));
            
            CreateLightUI(lightChooser);
            
        }

        private void CreateAdvancedUI()
        {
            MacGruber.Utils.RemoveUIElements(this, UIElements);
            MacGruber.Utils.RemoveUIElements(this, LightUIElements);
            UIElements.Clear();
            UIElements.Add(MacGruber.Utils.SetupButton(this, "Print Help", PrintHelp, true));
            UIElements.Add(MacGruber.Utils.SetupButton(this, "Back", CreateUI, false));
            
            UIElements.Add(CreateSlider(KeyInput.doubleTapTimeout, false));
            UIDynamicSlider slider = CreateSlider(IntensityRate, true);
            slider.valueFormat = "{0.000}";
            UIElements.Add(slider);
            UIElements.Add(CreateSlider(IntensityRate, true));
            
            UIElements.Add(CreateFilterablePopup(leftKeyChooser));
            UIElements.Add(CreateFilterablePopup(rightKeyChooser,true));
            UIElements.Add(CreateFilterablePopup(centerKeyChooser));
            UIElements.Add(CreateFilterablePopup(allKeyChooser,true));
            
            UIElements.Add(CreateFilterablePopup(mod0Chooser));
            UIElements.Add(CreateFilterablePopup(mod1Chooser,true));
            UIElements.Add(CreateFilterablePopup(mod2Chooser));
        }
        
        private void CreateMainHUD()
        {
            //.37f
            RectTransform panel = SuperController.singleton.GetAtomByUid("CoreControl").gameObject.transform.Find("WorldScaleAdjust/HUD/LowerHUDPivot/LowerHUDFlip/Scene Control Canvas/Panel").GetComponent<RectTransform>();
            versionRectTransform = panel.Find("VersionText").GetComponent<RectTransform>();
            versionRectTransform.anchorMax = new Vector2(.25f, 0f);
            headLights.ForEach(x => x. CreateHUDToggle(panel));
            
            toggleButtonTransform = manager.configurableScrollablePopupPrefab.transform.Find("Button") as RectTransform;
            toggleButtonTransform = Instantiate(toggleButtonTransform, panel);
            toggleButtonTransform.name = $"ToggleButton";
            toggleButtonTransform.anchorMax = new Vector2(.38f, 1);
            toggleButtonTransform.anchorMin = new Vector2(.25f, 0);
            toggleButtonTransform.offsetMax = new Vector2(0, -5f);
            toggleButtonTransform.offsetMin = new Vector2(0, 5f);
            Button button = toggleButtonTransform.GetComponent<Button>();
            Text buttonText = toggleButtonTransform.Find("Text").GetComponent<Text>();
            buttonText.text = "LightMeUp";
            button.onClick.AddListener(() => Toggle());
            toggleButtonTransform.gameObject.SetActive(false);
        }
    }
}